//**************************************************************************//
//* FRIDA: flexible rapid interactive data analysis                        *//
//* expr.cpp: evaluate generic expressions                                 *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://www.messen-und-deuten.de/frida                                  *//
//**************************************************************************//

#include <math.h>
#include <iostream>
#include <stdlib.h> // for boost/shared_ptr
#include <stdio.h>

#include <readln.h> // for macro parameters TODO

#include "mystd.h"  // for Integrate
#include "olf.h"
#include "mem.h"
#include "reg.h"
#include "func.h"
#include "var.h"
#include "expr.h"

extern bool allow_slow_conv; // TODO unify parameter treatment

#define DEBUG 0

//***************************************************************************//
//* class CContext                                                          *//
//***************************************************************************//

CContext::CContext( uint _k, uint _j, uint _i )
    : dim(_1), k(_k), j(_j), i(_i), want_error(false), nv(0), vt(0)
{
    if ( k==(uint)-1 )
        return; // not sure whether k=-1 makes sense

    POlo f = NOlm::MOM[k];
    if ( j==(uint)-1 ) {
        if ( f->nJ()==1 )
            j = 0;
    } else if ( j>=f->nJ() )
        throw "invalid evaluation context, j="+strg(j)+">=nJ="+strg(f->nJ());

    POld fd = P2D(f);
    if ( !fd ) {
        if ( i!=(uint)-1 )
            throw "invalid evaluation context, i set for curve file";
        return;
    }

    if ( j==(uint)-1 )
        return;
    uint ni = fd->nPts(j);
    if ( i==(uint)-1 ) {
        if ( ni==1 )
            i = 0;
    } else if ( i>=ni )
        throw "invalid evaluation context for j="+strg(j)+": i="+
            strg(i)+">=ni="+strg(ni);        
}

string CContext::context_info() const
{
    string out = "context: ";
    if      ( dim==_1 )
        out += "dim 1, ";
    else if ( dim==_VI )
        out += "dimVI, ";
    else if ( dim==_VT )
        out += "dimVT, ";
    out += "k = " + strg(k) + ", ";
    out += "j = " + strg(j) + ", ";
    out += "i = " + strg(i) + ", ";
    out += "nv = " + strg(nv) + ", ";
    if( vt )
        out += "size(vt) = " + strg(vt->size()) + ", ";
    else
        out += "no vt, ";
    out += "want_error = " + strg(want_error);
    return out;
}


//***************************************************************************//
//* class CResult                                                           *//
//***************************************************************************//

string CResult::result_info() const
{
    string out = "result: ";
    out += "vectorial = " + strg(vectorial) + ", ";
    out += "r = " + strg(r) + ", ";
    out += "size(v) = " + strg(v.size());
    return out;
}


//***************************************************************************//
//* class CRef (references like y[k,j,0] or p3[2] or ...)                   *//
//***************************************************************************//


//! Description of reference (for use in ??)

void CRef::set_coord( CCoord& ret, uint k_in ) const
{
    if     ( typ == _K ) {
        ret = CCoord("k","");
        return;
    } else if ( typ == _J ) {
        ret = CCoord("j","");
        return;
    } else if ( typ == _I ) {
        ret = CCoord("i","");
        return;
    }

    uint k;
    if (tk) {
        // get k from reference :
        CContext ctx( k_in );
        tk->tree_uival(&k, ctx);
        if (k>=NOlm::MOM.size())
            throw "set_coord: file index " + strg(k) + "invalid, maximum is " +
                strg(NOlm::MOM.size()-1);
    } else {
        k = k_in;
    }

    POlo f = NOlm::MOM[k];
    if (!f)
        throw "set_coord: ref where not refering to a file";

    if  ( typ == _Z ) {
        if (num>=f->ZCo.size())
            throw "set_coord: z" + strg(num) + " undefined";
        ret = f->ZCo[num];
    } else if  ( typ == _R ) {
        if (num>=f->RPar.size())
            throw "set_coord: r" + strg(num) + " undefined";
        ret = f->RPar[num].Co;
    } else if ( typ == _NJ ) {
        ret = CCoord("#spectra", "");
    } else {
        POld fd = P2D( f );
        POlc fc = P2C( f );
        if         ( fd ) {
            if       ( typ == _X )
                ret = fd->xco;
            else if  ( typ == _Y )
                ret = fd->yco;
            else if  ( typ == _DY )
                ret = CCoord( "d"+fd->yco.name, fd->yco.unit );
            else if  ( typ == _NI )
                ret = CCoord("#points", "");
            else
                throw "reference " + var_info() + " not allowed in data file";
        } else if ( fc ) {
            if     ( typ == _CP ) {
                if( num>=fc->nPar() )
                    throw "invalid reference p" + strg(num);
                ret = fc->PCo[num];
            } else if ( typ == _CQ )
                ret = CCoord("cq"+strg(num), ""); // fit quality indicator
            else if ( typ == _C )
                ret = CCoord(fc->expr,"");
            else
                throw "reference " + var_info() + " not allowed in curve file";
        } else
            throw "BUG: ref->set_coord unexpected else";
    }
}


//***************************************************************************//
//* CRef / evaluation                                                       *//
//***************************************************************************//

//! Evaluate reference in given context. Result returned in ret.

void CRef::ref_val( CResult& ret, const CContext& ctx ) const
{
    if ( (typ==_I || typ==_J || typ==_K) && ( ti || tj || tk ) )
        throw "invalid cross reference: index cannot be indexed";

    // Get k from context:
    CResult tmp;
    uint k;
    if ( tk ) {
        tk->tree_val( tmp, ctx );
        if ( tmp.vectorial ) {
            if ( typ==_K ) {
                ret.preset_v( ctx.nv );
                for( uint i=0; i<ctx.nv; ++i )
                    ret.v[i] = tmp.v[i];
                return;
            } else
                throw "TOO COMPLICATED: K must not depend on i";
        } else
            k = (uint) (tmp.r + 0.5);
    } else {
        k = ctx.k;
    }
    if ( k==(uint)-1 )
        throw "missing or invalid k";
    if ( k>=NOlm::MOM.size())
        throw "index k=" + strg(k) + " out of bounds";

    // Return k-dependent reference (unless already done above):
    if ( typ == _K ) {
        ret.set_r( k );
        return;
    }

    // Proceed with file reference f[k]:
    POlo f = NOlm::MOM[k];

    if ( typ==_NJ ) {
        ret.set_r( f->nJ() );
        return;
    } else if ( typ==_R ) {
        if ( num>= f->RPar.size() )
            throw "invalid reference " + var_info();
        ret.set_r( f->RPar[num].val );
        return;
    }

    // We need j:
    uint j;
    if ( tj ) {
        tj->tree_val( tmp, ctx );
        if ( tmp.vectorial ) {
            if ( typ==_J ) {
                ret.preset_v( ctx.nv );
                for( uint i=0; i<ctx.nv; ++i )
                    ret.v[i] = tmp.v[i];
                return;
            } else if ( typ==_Z ) {
                if ( num>= f->nZ() )
                    throw "invalid reference " + var_info();
                ret.preset_v( ctx.nv );
                for( uint i=0; i<ctx.nv; ++i ) {
                    j = (uint)( tmp.v[i]+0.5 );
                    if( j>=f->V.size() )
                        throw "j out of bounds";
                    ret.v[i] = f->V[j]->z[num];
                }
                return;
            } else
                throw "TOO COMPLICATED: J must not depend on i";
        } else
            j = (uint)( tmp.r + 0.5 );
    } else {
        j = ctx.j;
    }
    if ( j==-1 )
        throw "index j not defined";
    else if ( j>=f->nJ() )
        throw "index j=" + strg(j) + " out of bounds";

    // Return k-j-dependent reference (unless done above):
    if        ( typ == _J ) {
        ret.set_r( ctx.j );
        return;
    } else if ( typ == _Z ) {
        if ( num>= f->nZ() )
            throw "invalid reference " + var_info();
        ret.set_r( f->V[j]->z[num] );
        return;
    }

    // References for data file
    POld fd = P2D( f );
    if ( fd ) {
        PSpec sj = fd->VS(j);
        if        ( typ==_I || typ == _X || typ == _Y || typ == _DY ) {
            if ( typ==_DY && !sj->dy.size() )
                throw string( "dy not set" );
            if        ( ctx.dim==CContext::_1 ) {
                uint i;
                if ( ti ) {
                    ti->tree_uival( &i, ctx );
                } else {
                    i = ctx.i;
                    if ( i==(uint)-1 )
                        throw string( "missing index i" );
                }
                if( i>= sj->size() )
                    throw "i=" + strg(i) + " exceeds scan size" +
                        strg(sj->size());
                if       ( typ == _I ) {
                    ret.set_r( i );
                } else if ( typ == _X ) {
                    ret.set_r( sj->x[i] );
                } else if ( typ == _Y ) {
                    ret.set_r( sj->y[i] );
                    if ( ctx.want_error && sj->dy.size() )
                        ret.dr = sj->dy[i];
                } else if ( typ == _DY ) {
                    ret.set_r( sj->dy[i] );
                }
            } else if ( ctx.dim==CContext::_VI ) {
                if ( ti ) {
                    CContext myctx = ctx;
                    myctx.dim = CContext::_1;
                    ret.preset_v( ctx.nv );
                    if ( typ==_Y && ctx.want_error && sj->dy.size() )
                        ret.dv.resize( ctx.nv );
                    for( uint ii=0; ii<ctx.nv; ++ii ){
                        myctx.i = ii;
                        uint i;
                        ti->tree_uival( &i, myctx );
                        if( i >= sj->size() )
                            throw "i=" + strg(i) + " exceeds scan size" +
                                strg(sj->size());
                        if        ( typ == _I ) {
                            ret.v[ii] = i;
                        } else if ( typ == _X ) {
                            ret.v[ii] = sj->x[i];
                        } else if ( typ == _Y ) {
                            ret.v[ii] = sj->y[i];
                            if ( ret.dv.size() )
                                ret.dv[ii] = sj->dy[i];
                        } else if ( typ == _DY ) {
                            ret.v[ii] = sj->dy[i];
                        }
                    }
                } else {
                    if ( ctx.nv != sj->size() )
                        throw "ref_val " + var_info() + ": requested vec(" +
                            strg(ctx.nv) + "), found vec(" +
                            strg( sj->size() );
                    if        ( typ == _I ) {
                        ret.preset_v( ctx.nv );
                        for( uint i=0; i<ctx.nv; ++i )
                            ret.v[i] = i;
                    } else if ( typ == _X ) {
                        ret.set_v( sj->x );
                    } else if ( typ == _Y ) {
                        ret.set_v( sj->y );
                        if ( ctx.want_error )
                            ret.dv = sj->dy;
                    } else if ( typ == _DY ) {
                        ret.set_v( sj->dy );
                    }
                }
            } else
                throw string( "BUG: invalid context::dim" );
        } else if ( typ == _NI ) {
            ret.set_r( sj->size() );
        } else
            throw "invalid ref(" + var_info() + ") in data file";
        return;
    }

    // References for curve file:
    POlc fc = P2C( f );
    if ( fc ) {
        PCurve cj = fc->VC(j);
        if       ( typ == _CP ) {
            if ( num>= fc->nPar() )
                throw "invalid p ref(" + var_info() + ") in curve file";
            ret.set_r( cj->P[num] );
        } else if ( typ == _CQ ) {
            if ( num >= CCurve::mQuality )
                throw "invalid fm ref(" + var_info() + ") in curve file";
            ret.set_r( cj->Quality[num] );
        } else 
            throw "invalid ref(" + var_info() + ") in curve file";
        return;
    }
}

string CRef::ref_info() const
{
    string out = "ref: ";
    if ( tk )
        out += "tk[" + tk->tree_info() + "], ";
    else
        out += "tk=0, ";
    if ( tj )
        out += "tj[" + tj->tree_info() + "], ";
    else
        out += "tj=0, ";
    if ( ti )
        out += "ti[" + ti->tree_info() + "]";
    else
        out += "ti=0";
    return out;
}


//***************************************************************************//
//* class CTree (parsed expressions)                                        *//
//***************************************************************************//

//! Constructors.

CTree::CTree()
    : typ(_NOTREE), fun(0) {}

CTree::CTree( double v_in )
    : typ(_VAL), val(v_in), fun(0) {}

CTree::CTree( const class CFunc *f_in, PTree a0 )
    : typ(_OP), fun(f_in), narg(1)
{
    arg[0] = a0;
    if ( fun->narg!=1 )
        throw string( "BUG: CTree(1) with conflicting #arg" );
}

CTree::CTree( const class CFunc *f_in, PTree a0, PTree a1 )
    : typ(_OP), fun(f_in), narg(2)
{
    arg[0] = a0; arg[1] = a1;
    if ( fun->narg!=2 )
        throw string( "BUG: CTree(2) with conflicting #arg" );
}

CTree::CTree( const class CFunc *f_in, PTree a0, PTree a1, PTree a2 )
    : typ(_OP), fun(f_in), narg(3)
{
    arg[0] = a0; arg[1] = a1; arg[2] = a2;
    if ( fun->narg!=3 )
        throw string( "BUG: CTree(3) with conflicting #arg" );
}

CTree::CTree( PRef& _ref )
    : typ(_REF), fun(0), ref(_ref) {};

CTree::CTree( PRef& _ref, PTree a0 )
    : typ(_CEV), fun(0), ref(_ref)
{
    arg[0] = a0;
}

CTree::CTree( PRgr& _rgr )
    : typ(_RGR), rgr(_rgr) {};

CTree::CTree( char which )
{
    if        ( which=='t' ) { // function argument
        typ = _DUMMY;
    } else if ( which=='d' ) { // Dirac delta function centered at 0
        typ = _DIRAC;
        arg[0] = PTree( new CTree( 0.0 ) );
        narg = 1;
    } else
        throw string( "invalid tree type in char-constructor" );
}

CTree::CTree( char which, const PTree& a0 )
{
    if ( which=='d' ) { // Dirac delta function centered at a0
        typ = _DIRAC;
        if( a0->has_dummy() )
            throw string( "shift of Dirac delta must not contain t" );
        arg[0] = a0;
        narg = 1;
    } else
        throw string( "invalid tree type in char-constructor" );
}

CTree::CTree( char which, const PTree& a0, const PTree& a1 )
{
    if ( which=='c' ) { // convolution
        typ = _CONV;
        arg[0] = a0;
        arg[1] = a1;
        narg = 2;
    } else
        throw string( "invalid tree type in char-constructor" );
}


//***************************************************************************//
//* CTree / read out various properties                                     *//
//***************************************************************************//

//! How many curve parameters are contained in the tree?

uint CTree::npar() const
{
    uint np=0;
    npar_exec( &np );
    return np;
}


//! Implementation with pointer argument to allow recursion.

void CTree::npar_exec( uint *np ) const
{
    if      ( typ == _OP || typ == _CONV || typ == _DIRAC )
        for ( uint iarg=0; iarg<narg; ++iarg )
            arg[iarg]->npar_exec( np );
    else if ( typ == _VAL || typ == _DUMMY || typ == _CEV || typ == _RGR )
        ; // do nothing
    else if ( typ == _REF ) {
        if ( ref->typ==CRef::_CP )
            *np = max (*np, ref->num+1);
    } else
        throw string( "BUG: npar(UNDEF)" );
}


//! Does the result depend on k?

bool CTree::k_dependent() const
{
    bool kd=false;
    kdep_exec( &kd );
    return kd;
}


//! Implementation with pointer argument to allow recursion.

void CTree::kdep_exec( bool *kd ) const
{
    if      ( typ == _OP || typ == _CONV || typ == _DIRAC )
        for ( uint iarg=0; iarg<narg; ++iarg )
            arg[iarg]->kdep_exec( kd );
    else if ( typ == _VAL || typ == _DUMMY || typ == _CEV || typ == _RGR )
        ; // do nothing
    else if ( typ == _REF ) {
        *kd = true;
    } else
        throw string( "BUG: kdep(UNDEF)" );
}


//! Does expression contain dummy argument t ?

bool CTree::has_dummy() const
{
    switch ( typ ) {
    case _OP:
        for ( uint iarg=0; iarg<narg; ++iarg )
            if( arg[iarg]->has_dummy() )
                return true;
        break;
    case _DIRAC:
        return true; // implicit
    case _CONV:
        return arg[0]->has_dummy();
    case _DUMMY:
        return true;
    default:
        ;
    }
    return false;
}


//! Compose coordinate names and units according to the tree.

void CTree::set_coord( CCoord& ret, uint k ) const
{
    CCoord r[maxarg];
    switch ( typ ) {
    case _OP:
        for ( uint iarg=0; iarg<narg; ++iarg )
            arg[iarg]->set_coord( r[iarg], k );
        if     ( narg==1 )
            ret = fun->coord( r+0);
        else if ( narg==2 ) 
            ret = fun->coord( r+0, r+1 );
        else if ( narg==3 ) 
            ret = fun->coord( r+0, r+1, r+2 );
        break;
    case _CONV:
        arg[0]->set_coord( ret, k );
        break;
    case _DIRAC:
        ret = CCoord("dirac", "");
        break;
    case _VAL:
        ret = CCoord(strg(val), "");
        break;
    case _REF:
        ref->set_coord( ret, k );
        break;
    case _CEV:
        ret = CCoord( "c", "" ); // primitiver gehts nicht
        break;
    case _RGR:
        ret = CCoord( "$", "" ); // primitiver gehts nicht
        break;
    default:
        throw string( "BUG: tree::set_coord(tree_typ=UNDEF)" );
    }
}


//! Tree -> expression.

string CTree::tree_info() const
{
    string ret;
    switch ( typ ) {
    case _OP:
        ret = fun->txt + "(";
        for ( uint iarg=0; iarg<narg; ++iarg )
            ret += arg[iarg]->tree_info() + (iarg<narg-1 ? "," : ")" );
        return ret;
    case _CONV:
        return "conv(" + arg[0]->tree_info() + ")";
    case _DIRAC:
        return "dirac(" + arg[0]->tree_info() + ")";
    case _VAL:
        return "val[" + strg(val) + "]";
    case _REF:
        return "ref[" + ref->var_info() + "]";
    case _CEV:
        return "c[" + ref->var_info() + "]";
    case _RGR:
        return "$";
    case _DUMMY:
        return "farg";
    default:
        throw string( "BUG: tree_info: unexpected tree type" );
    }
}


//***************************************************************************//
//* CTree / low-level tree evaluation                                       *//
//***************************************************************************//

//! Evaluate tree and convert result to integer, for use as index.

void CTree::tree_uival( uint *ret, const CContext& ctx ) const
{
    CResult val;
    tree_val( val, ctx );
    // since we have no integer data type, convert real by rounding
    *ret = (uint) ( val.r + 0.5);
}


//! Evaluate tree. Private. User calls pass through API implemented below.

void CTree::tree_val( CResult& ret, const CContext& ctx ) const
{
    switch ( typ ) {
    case _OP:
        tree_op( ret, ctx );
        break;
    case _CONV: case _DIRAC:
        tree_conv( ret, ctx );
        break;
    case _VAL:
        ret.set_r( val );
        break;
    case _DUMMY:
        if ( !(ctx.vt) )
            throw "BUG: tree_val with *vt=0";
        ret.set_v( *(ctx.vt) );
        break;
    case _REF:
        ref->ref_val( ret, ctx );
        break;
    case _RGR:
        ret.set_r( NReg::lastresult );
        break;
    case _CEV:
        tree_cev( ret, ctx );
        break;
    default:
        throw "BUG: unexpected tree type";
    }
}


//! A function of 1,2,.. arguments (including operators like +,-,.. ).

void CTree::tree_op( CResult& ret, const CContext& ctx ) const
{
    CResult a[maxarg];
    // evaluate arguments
    for ( uint iarg=0; iarg<narg; ++iarg )
        arg[iarg]->tree_val( a[iarg], ctx );
    // scalar arguments ?
    bool is_scalar = true;
    for ( uint iarg=0; iarg<narg; ++iarg ) {
        if ( !a[iarg].vectorial )
            continue;
        is_scalar = false;
    }
    // some input has errors ?
    bool err_input = false;
    for ( uint iarg=0; iarg<narg; ++iarg ) {
        if ( a[iarg].has_err() ){
            err_input = true;
            break;
        }
    }
    // size of vectorial arguments
    int n=0;
    if( !is_scalar ) {
        for ( uint iarg=0; iarg<narg; ++iarg ) {
            if ( !a[iarg].vectorial )
                continue;
            if ( n==0 )
                n = a[iarg].v.size();
            else if ( a[iarg].v.size() != n )
                throw "vector arguments have different size";
        }
        ret.preset_v( n );
    }
    // now evaluate the function
    if       ( narg==1 ) {
        if( ctx.want_error && fun->d1 && err_input ) {
            if ( is_scalar ) {
                (*(fun->d1))( ret.r, ret.dr, a[0].r, a[0].dr );
            } else {
                ret.dv.resize( n );
                for ( int i=0; i<n; ++i )
                    (*(fun->d1))( ret.v[i], ret.dv[i], a[0].v[i], a[0].dv[i] );
            }
        } else {
            if ( is_scalar )
                ret.set_r( (*(fun->f1))( a[0].r ) );
            else
                for ( int i=0; i<n; ++i )
                    ret.v[i] = (*(fun->f1))( a[0].v[i] );
        }
    } else if ( narg==2 ) {
        if( ctx.want_error && ( ( fun->d2 && err_input ) ||
                                fun->txt=="+-" ) ) {
            if ( is_scalar ) {
                (*(fun->d2))( ret.r, ret.dr, a[0].r, a[0].dr, a[1].r, a[1].dr );
            } else {
                ret.dv.resize( n );
                for ( int i=0; i<n; ++i )
                    (*(fun->d2))( ret.v[i], ret.dv[i],
                                  a[0].to_r(i), a[0].to_dr(i),
                                  a[1].to_r(i), a[1].to_dr(i) );
            }
        } else {
            if       ( !a[0].vectorial && !a[1].vectorial )
                ret.set_r( (*(fun->f2))( a[0].r, a[1].r ) );
            else if  ( !a[0].vectorial && a[1].vectorial ) {
                for ( int i=0; i<n; ++i )
                    ret.v[i] = (*(fun->f2))( a[0].r, a[1].v[i] );
            } else if ( a[0].vectorial && !a[1].vectorial ) {
                for ( int i=0; i<n; ++i )
                    ret.v[i] = (*(fun->f2))( a[0].v[i], a[1].r );
            } else if ( a[0].vectorial && a[1].vectorial ) {
                for ( int i=0; i<n; ++i )
                    ret.v[i] = (*(fun->f2))( a[0].v[i], a[1].v[i] );
            }
        }
    } else if ( narg==3 ) {
        if( ctx.want_error && fun->d3 && err_input ) {
            if ( is_scalar ) {
                (*(fun->d3))( ret.r, ret.dr, a[0].r, a[0].dr, a[1].r, a[1].dr,
                              a[2].r, a[2].dr );
            } else {
                ret.dv.resize( n );
                for ( int i=0; i<n; ++i )
                    (*(fun->d3))(
                        ret.v[i], ret.dv[i],
                        a[0].to_r(i), a[0].to_dr(i),
                        a[1].to_r(i), a[1].to_dr(i),
                        a[2].to_r(i), a[2].to_dr(i) );
            }
        } else {
            if ( is_scalar ) 
                ret.set_r( (*(fun->f3))( a[0].r, a[1].r, a[2].r ) );
            else {
                for ( int i=0; i<n; ++i )
                    ret.v[i] = (*(fun->f3))(
                        a[0].to_r(i), a[1].to_r(i), a[2].to_r(i) );
            }
        }
    }
    if( DEBUG ) {
        cout << "tree_op:\n";
        for ( uint iarg=0; iarg<narg; ++iarg )
            cout << "  arg: " << a[iarg].result_info() << "\n";
        cout << "  res: " << ret.result_info() << "\n";
    }
}


//! A curve evaluation: fc[k,j](arg).

void CTree::tree_cev( CResult& ret, const CContext& ctx ) const
{
    // get k from context:
    uint k;
    if ( ref->tk ) {
        CResult tmp;
        ref->tk->tree_val( tmp, ctx );
        if ( tmp.vectorial )
            throw "TOO COMPLICATED: in curves, K must not depend on i";
        k = (uint)( tmp.r + 0.5 );
    } else {
        k = ctx.k;
    }
    if ( k==(uint)-1 )
        throw "missing or invalid k";
    else if ( k>=NOlm::MOM.size())
        throw "index k=" + strg(k) + " out of bounds";

    // k points to a curve:
    POlc fc = P2C( NOlm::MOM[k] );
    if( !fc )
        throw "file " + strg(k) + " is not a curve";

    // get j from context:
    uint j;
    if ( ref->tj ) {
        CResult tmp;
        ref->tj->tree_val( tmp, ctx );
        if ( tmp.vectorial )
            throw "TOO COMPLICATED: in curves, J must not depend on i";
        j = (uint)( tmp.r + 0.5 );
    } else {
        j = ctx.j;
    }
    if ( j==(uint)-1 ) {
        if ( fc->nJ()==1 )
            j = 0;
        else
            throw "missing or invalid j";
    }

    // evaluate and pack function argument:
    CResult farg;
    arg[0]->tree_val( farg, ctx );

    // evaluate the curve:
    if ( farg.vectorial ) {
        ret.preset_v( farg.v.size() );
        fc->curve_val_vec( &(ret.v), farg.v, k, j );
    } else
        ret.set_r( fc->curve_val_sca( farg.r, k, j ) );
}


//***************************************************************************//
//* CTree / public interface to tree evaluation                             *//
//***************************************************************************//

//! Evaluate tree. Public API.

void CTree::tree_point_val(
    double *ret, double *dret, uint k, uint j, uint i ) const
{
    CContext ctx( k, j, i );
    ctx.want_error = (bool)dret;
    CResult val;
    tree_val( val, ctx );
    if ( val.vectorial )
        throw "BUG: tree_point_val did not get scalar result";
    *ret = val.r;
    if( dret )
        *dret = val.dr;
}


//! Evaluate tree. Public API.

void CTree::tree_vec_val(
    vector<double> *ret, vector<double> *dret, uint k, uint j ) const
{
    CResult val;
    CContext ctx( k, j );
    uint nret = ret->size();
    ctx.request_VI( nret );
    ctx.want_error = (bool)dret;
    tree_val( val, ctx );
    if ( val.vectorial ){
        *ret = val.v;
        if( dret )
            *dret = val.dv;
    } else {
        ret->assign( nret, val.r );
        if( dret && val.dr )
            dret->assign( nret, val.dr );
    }
}


//***************************************************************************//
//* CTree / convolution evaluation                                          *//
//***************************************************************************//

//! Return kconc and jconv, according to fc[k,j]->kconv.

void CTree::setConv( uint& kconv, uint& jconv, uint k, uint j )
{
    // set kconv for given k:
    if ( k >= NOlm::MOM.size() )
        throw "BUG: setConv called with invalid k";
    POlc fc = P2C( NOlm::MOM[k] );
    if( !fc )
        throw "k refers to data, curve expected";
    kconv = fc->kconv;
    if( kconv==(uint)-1 ) { // no convolution
        jconv = -1; // should not be needed
        return;
    } else if ( kconv >= NOlm::MOM.size() )
        throw "invalid convolution file index " + strg(kconv);
    POld fconv = P2D( NOlm::MOM[kconv] );
    if( !fconv )
        throw "cannot convolute with curve";

    // set jconv for given j:
    if ( j>=fc->nJ() )
        throw "BUG: setConv called with invalid j";
    if      ( fconv->nJ()==1 )       // nJ:1  correspondence
        jconv = 0;
    else if ( fconv->nJ()==fc->nJ() ) // nJ:nJ correspondence
        jconv = j;
    else
        throw "cannot convolute file " + strg(k) + " with resolution " +
            strg(kconv) + ": number of spectra does not match";
}


//! Convolution context. Declared locally because only needed in tree_conv.

typedef struct {
    PSpec sv;
    uint nv;
    double t;
    CContext cvctx;
    CResult res;
    PTree arg;
} ConvDatTyp;


//! Call-back for integration. Only needed in tree_conv.

double tree_conv_eval( double tt, void *data )
{
    ConvDatTyp *mydata = (ConvDatTyp*) data;
    if( tt < mydata->sv->x[0] || tt > mydata->sv->x[mydata->nv-1] )
        throw string( "BUG: gsl requests invalid integration point" );
    vector<double> t1(1);
    t1[0] = mydata->t - tt;
    mydata->cvctx.vt = &t1;
    mydata->arg->tree_val( mydata->res, mydata->cvctx );
    return mydata->res.to_r(0) * mydata->sv->intpol( tt );
}


//! Convolution or Dirac function: conv(f, shift), dirac(shift).

void CTree::tree_conv( CResult& ret, const CContext& ctx ) const
{
    // Return values are requested for n values of _DUMMY ( = t ):
    if ( !(ctx.vt) )
        throw string( "BUG: treeval(conv,dirac) with *vt=0" );
    uint n = ctx.vt->size();
    ret.preset_v( n );
    ret.dv.clear();
    // Resolution is given by context 'ctx':
    uint kconv, jconv;
    setConv( kconv, jconv, ctx.k, ctx.j );
    double conv_norm; // normalization constant (integral of resolution)
    double conv_step; // step in x/t, if equidistant
    PSpec sv;
    int nv; /* declaring this as unsigned lead to wrong wrapping */
    if( kconv!=-1 ){
        // Check whether return grid is equidistant
        if( n>1 ){
            conv_step = ((*(ctx.vt))[n-1]-(*(ctx.vt))[0]) / (n-1);
            if( conv_step<0 )
                conv_step = 0;
            else {
                for( int i=0; i<n-1; ++i ) {
                    if( fabs((*(ctx.vt))[i+1]-(*(ctx.vt))[i]
                             -conv_step)>1e-8*conv_step ){
                        conv_step = 0; // not equidistant
                        break;
                    }
                }
            }
        } else 
            conv_step = 0;
        // Preset fv,sv,nv.
        POld fv = P2D( NOlm::MOM[kconv] );
        if ( !fv )
            throw "Resolution " + strg(kconv) + "is not a data file";
        sv = fv->VS(jconv);
        nv = sv->size();
        if ( nv < 2 )
            throw "Resolution " + strg(kconv) + ", spectrum " +
                strg(jconv) + ", has only " + strg(nv) + " points";
        // Calculate normalization integral.
        // At this occasion, assert that x is ordered,
        conv_norm = 0;
        // special treatment for first and last step:
        double dx = sv->x[1] - sv->x[0];
        if ( dx <= 0 )
            throw "Resolution " + strg(kconv) + ", spectrum " +
                strg(jconv) + ", not ordered in x";
        conv_norm += sv->y[0] * dx;
        dx = sv->x[nv-1] - sv->x[nv-2];
        if ( dx <= 0 )
            throw "Resolution " + strg(kconv) + ", spectrum " +
                strg(jconv) + ", not ordered in x";
        conv_norm += sv->y[nv-1] * dx;
        // the other steps:
        for( int i=1; i<nv-1; ++i ) {
            dx = ( sv->x[i+1] - sv->x[i-1] ) / 2;
            if ( dx <= 0 )
                throw "Resolution " + strg(kconv) + ", spectrum " +
                    strg(jconv) + ", not ordered in x";
            conv_norm += sv->y[i] * dx;
        }
        // check result:
        if ( conv_norm <= 0 )
            throw "Resolution " + strg(kconv) + ", spectrum " +
                strg(jconv) + ", has non-positive norm " + strg(conv_norm);
        // Check whether x is equidistant
        if( conv_step ){
            for( int i=0; i<nv-1; ++i ) {
                if( fabs(sv->x[i+1]-sv->x[i]-conv_step)>1e-8*conv_step ){
                    conv_step = 0; // not equidistant
                    break;
                }
            }
        }
        if( !conv_step && !allow_slow_conv )
            throw string(
                "not equidistant and slow convolution not allowed" );
    }
    // The actual computation is rather different for _CONV and _DIRAC:
    if        ( typ == _CONV ) {
        if( narg!=2 )
            throw string( "BUG: convolution requires 2 args" );
        CResult arg_shift;
        arg[1]->tree_val( arg_shift, ctx );
        if ( arg_shift.vectorial )
            throw string( "2nd arg of conv must be scalar shift" );
        double shift = arg_shift.r;
        if ( kconv==-1 ) {
            // Copy convolutand (= unconvoluted theory).
            CResult theory;
            // unconvoluted theory given by arg[0]:
            arg[0]->tree_val( theory, ctx );
            for ( int i=0; i<n; ++i )
                ret.v[i] = theory.to_r(i);
        } else if ( NRead::iofmac("\\cv_mode")==-1 ) {
            CResult theory;
            CContext cvctx(ctx);
            vector<double> tt;
            cvctx.vt = &tt;
            // Convolute theory with resolution.
            if( conv_step ){
                // Accelerated version with equidistant grids 
                int ntt = n + nv - 1;
                tt.resize(ntt);
                for ( int ii=0; ii<ntt; ++ii ) // ii=(nv-1)+i-iv
                    tt[ii] = (*(ctx.vt))[0] - sv->x[0] - shift +
                        (ii-(nv-1))*conv_step;
                arg[0]->tree_val( theory, cvctx );
                for ( int i=0; i<n; ++i ){
                    ret.v[i] = 0;
                    for ( int iv=0; iv<nv; ++iv )
                        ret.v[i] += theory.to_r(nv-1+i-iv) * sv->y[iv];
                    ret.v[i] *= conv_step / conv_norm;
                }
            } else {
                // Simplest version: non-equidistant x_out and x_res.
                for ( int i=0; i<n; ++i )
                    ret.v[i] = 0;
                tt.resize(n);
                for ( int iv=0; iv<nv; ++iv ) {
                    double dx;
                    if     ( iv==0 )
                        dx = sv->x[1] - sv->x[0];
                    else if( iv==nv-1 )
                        dx = sv->x[nv-1] - sv->x[nv-2];
                    else
                        dx = ( sv->x[iv+1] - sv->x[iv-1] ) / 2;
                    for ( int i=0; i<n; ++i )
                        tt[i] = (*(ctx.vt))[i] - sv->x[iv] - shift;
                    arg[0]->tree_val( theory, cvctx );
                    for ( int i=0; i<n; ++i )
                        ret.v[i] += theory.to_r(i) * sv->y[iv] *
                            dx / conv_norm;
                }
            }
        } else {
            // Convolute theory with resolution.
            // Generic integration.
            ConvDatTyp conv_data;
            conv_data.sv = sv;
            conv_data.nv = nv;
            conv_data.arg = arg[0];
            conv_data.cvctx = ctx;
            for ( int i=0; i<n; ++i ) {
                conv_data.t = (*(ctx.vt))[i] - shift;
                ret.v[i] = mystd::Integrate(
                    &tree_conv_eval,
                    (void *) &conv_data,
                    sv->x[0], sv->x[nv-1],
                    NRead::iofmac( "\\cv_mode" ),
                    NRead::dofmac( "\\cv_epsabs" ),
                    NRead::dofmac( "\\cv_epsrel" ) );
                ret.v[i] /= conv_norm;
            }
        }

    } else if ( typ == _DIRAC ) {
        CResult arg_shift;
        // center of delta function given by arg[0]:
        arg[0]->tree_val( arg_shift, ctx );
        if ( arg_shift.vectorial )
            throw string( "dirac function requires scalar argument" );
        if ( kconv==-1 ) {
            // Naive representation of delta function.
            for ( int i=0; i<n; ++i )
                ret.v[i] = fabs((*(ctx.vt))[i]-arg_shift.to_r(i))<1e-2 ?
                    1e2 : 0;
        } else {
            // Delta function copies resolution.
            vector<double> tt(n), yy(n);
            // shift in t
            for ( int i=0; i<n; ++i )
                tt[i] = (*(ctx.vt))[i]-arg_shift.r;
            // interpolation
            sv->intpol( tt, yy );
            // normalization
            for ( int i=0; i<n; ++i )
                ret.v[i] = yy[i] / conv_norm;
        }
    }
}
